@@ -27,6 +27,24 @@ class DataProvider():
             self.fileref = open(self.filename, mode)
         except(Exception):
             raise LaspyException("Error opening file")
+    def get_point_map(self, informat):
+        '''Get point map is used to build and return a numpy frombuffer view of the mmapped data, 
+        using a valid laspy.util.Format instance for the desired point format. This method is used 
+        to provide access to extra_bytes even when dimensions have been explicitly defined via an 
+        extra_bytes VLR record.'''
+        if type(self._mmap) == bool:
+            self.map() 
+        self.pointfmt = np.dtype([("point", zip([x.name for x in informat.specs],
+                                [x.np_fmt for x in informat.specs]))]) 
+        if not self.manager.header.version in ("1.3", "1.4"): 
+            _pmap = np.frombuffer(self._mmap, self.pointfmt, 
+                        offset = self.manager.header.data_offset)
+        else:  
+            _pmap = np.frombuffer(self._mmap, self.pointfmt, 
+                        offset = self.manager.header.data_offset,
+                        count = self.manager.header.point_records_count)
+        return(_pmap)
+
 
     def point_map(self):
         '''Create the numpy point map based on the point format.'''   
@@ -161,6 +179,20 @@ class FileManager():
             self.set_vlrs(vlrs)
         if evlrs != False:
             self.set_evlrs(vlrs)
+
+        # If extra-bytes descriptions exist in VLRs, use them.
+        eb_vlrs = [x for x in self.vlrs if x.type == 1]
+        eb_vlrs.extend([x for x in self.evlrs if x.type == 1])
+        self.extra_dimensions = []
+        if len(eb_vlrs) > 1:
+            raise LaspyException("Only one ExtraBytes VLR currently allowed.")
+        elif len(eb_vlrs) == 1:
+            self.naive_point_format = self.point_format
+            self.extra_dimensions = eb_vlrs[0].extra_dimensions
+            new_pt_fmt = Format(self.point_format.fmt, extradims 
+                    = self.extra_dimensions)
+            self.point_format = new_pt_fmt
+            self.data_provider.remap(point_map = True)
         return
 
     def setup_write(self,header, vlrs, evlrs):
@@ -172,7 +204,6 @@ class FileManager():
         self.header_format = header.format 
         self._header = header
         self.header = HeaderManager(header = header, reader = self)
-        
 
         self.initialize_file_padding(vlrs)
 
@@ -190,7 +221,19 @@ class FileManager():
         if self._header.created_year == 0:
             self.header.date = datetime.datetime.now() 
         self.populate_vlrs()
-        #self.populate_evlrs()
+        self.populate_evlrs()
+        # If extra-bytes descriptions exist in VLRs, use them.
+        eb_vlrs = [x for x in self.vlrs if x.type == 1]
+        eb_vlrs.extend([x for x in self.evlrs if x.type == 1])
+        self.extra_dimensions = []
+        if len(eb_vlrs) > 1:
+            raise LaspyException("Only one ExtraBytes VLR currently allowed.")
+        elif len(eb_vlrs) == 1:
+            self.naive_point_format = self.point_format
+            self.extra_dimensions = eb_vlrs[0].extra_dimensions
+            new_pt_fmt = Format(self.point_format.fmt, extradims 
+                    = self.extra_dimensions)
+            self.point_format = new_pt_fmt
         return
 
     def correct_rec_len(self):
@@ -344,12 +387,21 @@ class FileManager():
 
     def get_vlrs(self):
         '''Populate and return list of :obj:`laspy.header.VLR` objects`.'''
-        self.populate_vlrs()
-        return(self.vlrs)
-   
+        try:
+            return(self.vlrs)
+        except:
+            self.populate_vlrs()
+            return(self.vlrs)
+    
+    def push_vlrs(self):
+        self.set_vlrs(self.vlrs)
+
     def get_evlrs(self):
-        self.populate_evlrs()
-        return(self.evlrs)
+        try:
+            return(self.evlrs)
+        except:
+            self.populate_evlrs()
+            return(self.evlrs)
 
     def get_padding(self):
         '''Return the padding between the end of the VLRs and the beginning of
@@ -683,6 +735,9 @@ class FileManager():
     def get_extra_bytes(self):
         if "extra_bytes" in self.point_format.lookup.keys():
             return(self.get_dimension("extra_bytes"))
+        elif self.extra_dimensions != []:
+            newmap = self.data_provider.get_point_map(self.naive_point_format) 
+            return(newmap["point"]["extra_bytes"])
         else:
             raise LaspyException("Extra bytes not present in record")
 
@@ -804,15 +859,40 @@ class Writer(FileManager):
             self.data_provider.open("r+b")
             self.data_provider.map()
             self.data_provider.point_map()
-        elif self.mode == "w" and not self.has_point_records:
+            self.populate_vlrs()
+        elif self.mode == "w" and not self.has_point_records: 
+            self.set_header_property("num_variable_len_recs", len(value))
             self.seek(self.header.header_size, rel = False)
-            for vlr in value: 
+            for vlr in value:
                 self.data_provider._mmap.write(vlr.to_byte_string())
+            self.populate_vlrs()
             return
         else:
-            raise(LaspyException("set_vlrs requires the file to be opened in a " + 
-                "write mode, and must be performed before point information is provided." + 
-                "Try closing the file and opening it in rw mode. "))
+            current_size = self.data_provider._mmap.size()
+            current_padding = self.get_padding()
+            old_offset = self.header.data_offset
+            new_offset = current_padding + self.header.header_size + sum([len(x) for x in value])
+            self.set_header_property("data_offset", new_offset)
+            self.set_header_property("num_variable_len_recs", len(value))
+            self.data_provider.fileref.seek(0, 0)
+            dat_part_1 = self.data_provider.fileref.read(self.header.header_size)
+            self.data_provider.fileref.seek(old_offset, 0)
+            dat_part_2 = self.data_provider.fileref.read(current_size - old_offset)
+            # Manually Close:
+            self.data_provider.close(flush=False)
+            self.data_provider.open("w+b")
+            self.data_provider.fileref.write(dat_part_1)
+            for vlr in value:
+                byte_string = vlr.to_byte_string()
+                self.data_provider.fileref.write(byte_string)
+            self.data_provider.fileref.write("\x00"*current_padding)
+            self.data_provider.fileref.write(dat_part_2)
+            self.data_provider.fileref.close()
+            self.data_provider.open("r+b")
+            self.data_provider.map()
+            self.data_provider.point_map()
+            self.populate_vlrs() 
+
 
     def set_padding(self, value):
         '''Set the padding between end of VLRs and beginning of point data'''
@@ -1360,5 +1440,9 @@ class Writer(FileManager):
         '''Wrapper for set_dimension("extra_bytes")'''
         if "extra_bytes" in self.point_format.lookup.keys():
             self.set_dimension("extra_bytes", extra_bytes)
+        elif self.extra_dimensions != []:
+            newmap = self.data_provider.get_point_map(self.naive_point_format) 
+            newmap["point"]["extra_bytes"] = extra_bytes
         else:
             raise LaspyException("Extra bytes not present in point format. Try creating a new file with an extended point record length.")
+
